import 'package:bird/base.dart';
import 'package:bird/fp.dart';
import 'package:bird/signal_wave.dart';
import 'package:meta/meta.dart';

/// Base interfaces for classes that know how to
/// - set a value
/// - get that value synchronously
/// - subscribe to a wave to receive new values.
abstract class PureSea<V> {
  Wave<V> get wave;

  void add(V v);

  V get value;
}

abstract class Sea<V> extends Wave<V>
    with WaveType<V>, WaveOtherMixin<V>, WaveOperators<V>
    implements PureSea<V> {

  factory Sea(V v) => _Sea(v);

  Sea.wave(WaveType<V> wave) : super(wave);

  void change(V Function(V) action) {
    add(action(value));
  }

  SeaLens<V, R> seaThen<R>(Glass<V, R, V, R, V> glass) {
    return SeaLens.rawLens(this, glass);
  }

  Sea<R> transform<R>(Glass<V, R, V, R, V> glass) {
    return _AnonymousSea(
      this.map(glass.get),
          () => glass.get(value),
          (val) => this.add(glass.set(this.value, val)),
    );
  }
}

/// A SeaLens has access to a source [Sea] and is also a lens.
class SeaLens<S, T> extends ConcreteGlass<void, S, void, T, void> {
  final Sea<S> sourceSea;

  final Glass<S, T, S, T, S> lens;

  const SeaLens.raw(this.sourceSea, this.lens, S Function(void) get,
      void Function(void, T) set) : super(get, set);

  factory SeaLens.rawLens(Sea<S> sourceSea, Glass<S, T, S, T, S> lens) {
    return SeaLens.raw(
      sourceSea,
      lens,
          (_) => sourceSea.value,
          (_, T t) => sourceSea.add(lens.set(sourceSea.value, t)),
    );
  }

  SeaLens<S, Q> seaThen<Q>(Glass<T, Q, T, Q, T> glass) {
    return SeaLens.rawLens(sourceSea, lens.then(glass));
  }

  SeaMSLens<S, Q, F> seaThenRM<Q, F extends Monad<F, dynamic>>(
      Glass<T, Q, T, Q, Kind<F, T>> glass) {
    return SeaMSLens.rawLens(sourceSea, lens.thenRM(glass));
  }
}

/// A SeaLens has access to a source [Sea] and is also a lens.
class SeaMSLens<S, T, F extends Monad<F, dynamic>>
    extends ConcreteGlass<void, S, void, T, void> {
  final Sea<S> sourceSea;

  final Glass<S, T, S, T, Kind<F, S>> lens;

  const SeaMSLens.raw(this.sourceSea, this.lens, S Function(void) get,
      void Function(void, T) set) : super(get, set);

  factory SeaMSLens.rawLens(Sea<S> sourceSea,
      Glass<S, T, S, T, Kind<F, S>> lens) {
    return SeaMSLens.raw(
      sourceSea,
      lens,
          (_) => sourceSea.value,
          (_, T t) {
        final Kind<Monad, S> functor = lens.set(sourceSea.value, t);
        (functor as Monad<F, S>).map(sourceSea.add);
      },
    );
  }

  SeaMSLens<S, Q, F> seaThenRM<Q>(Glass<T, Q, T, Q, Kind<F, T>> glass) {
    return SeaMSLens.rawLens(sourceSea, lens.thenRM<Q, F>(glass));
  }
}

class _AnonymousSea<V> extends Sea<V> {
  @protected
  final void Function(V) _add;

  @override
  void add(V v) => _add(v);

  final V Function() __value;

  @override
  V get value => __value();

  @override
  final Wave<V> wave;

  _AnonymousSea(this.wave, this.__value, this._add) : super.wave(wave);
}

/// A [Wave] with a synchronously available and changeable value.
/// Meant to reduce Signal boilerplate.
///
/// It has to be disposed.
///
/// It doesn't implement Sink to suppress the close-sinks
/// lint warning that Signals have.
///
/// A [Sea] is also a [Wave] so that it is backwards compatible with consumers
/// expecting a [Wave].
///
/// Add is only meant to be called inside of HookBlocs. That's why it's set
/// as protected.
class _Sea<V> extends Sea<V> {
  final Future<void> Function() _close;

  @override
  final Wave<V> wave;

  /// .add is supposed to be called in HookBloc subclasses never outside.
  /// You're supposed to add this line to the top of your HookBloc.
  /// You might also want to ignore type declarations because they get inferred.
  // ignore_for_file: invalid_use_of_protected_member
  // ignore_for_file: type_annotate_public_apis
  @protected
  final void Function(V) _add;

  @override
  void add(V v) => _add(v);

  final V Function() __value;

  @override
  V get value => __value();

  factory _Sea(V s) {
    final signal = Signal(s);
    return _Sea._create(
          () async => signal.close(),
      signal.wave,
      signal.add,
          () => signal.value,
    );
  }

  _Sea._create(this._close, this.wave, this._add, this.__value)
      : super.wave(wave);

  /// Use this in combination with [SeaHookBlocExtension.d] to dispose this
  /// [Sea] and hide the implementation details.
  Sea<V> operator &(
      _Sea<V> Function(_Sea<V> self, void Function() dispose) action) {
    return action(this, _close);
  }
}

/// Experimental High level HookBloc extensions for boilerplate free Blocs.
extension SeaHookBlocExtension on HookBloc {
  Sea<T> d<T>(_Sea<T> t, void Function() dispose) {
    onDispose.add(() async => dispose());
    return t;
  }
}

extension SeaExtension<T> on T {
  /// Use this inside of [HookBloc]s as a replacement for [HookBloc.disposeSink].
  Sea<T> get $ {
    final sea = _Sea(this);
    HookBloc.disposeEffect(sea, (_Sea<T> a) async => a._close());
    return sea;
  }

  /// This is meant to be use when initialized inside of constructors.
  /// Use the & operator and [d] extension method on [HookBloc]s to dispose
  /// this [Sea].
  ///
  /// ```
  /// final a = 1.sea & d;
  /// ```
  ///
  _Sea<T> get $$ {
    return _Sea(this);
  }
}
